package it.cavallium.strangedb.database;

import it.cavallium.strangedb.database.references.DatabaseReferencesIO;
import it.cavallium.strangedb.database.references.DatabaseReferencesMetadata;
import it.cavallium.strangedb.database.references.ReferenceInfo;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardCopyOption;

import static it.cavallium.strangedb.database.references.DatabaseReferencesMetadata.NONEXISTENT_REFERENCE_INFO;

public class DatabaseCore implements IDatabase {

	private final DatabaseFileIO fileIO;
	protected final DatabaseReferencesIO referencesIO;
	protected final DatabaseReferencesMetadata referencesMetadata;
	protected final Path dataFile;
	protected final Path referencesMetaFile;
	protected volatile boolean closed;

	public DatabaseCore(Path dataFile, Path referencesMetaFile) throws IOException {
		if (Files.notExists(dataFile)) {
			Files.createFile(dataFile);
		}
		if (Files.notExists(referencesMetaFile)) {
			Files.createFile(referencesMetaFile);
		}
		this.dataFile = dataFile;
		this.referencesMetaFile = referencesMetaFile;
		this.fileIO = new DatabaseFileIO(dataFile);
		this.referencesMetadata = new DatabaseReferencesMetadata(referencesMetaFile);
		this.referencesIO = new DatabaseReferencesIO(fileIO, referencesMetadata);
	}

	@Override
	public synchronized void close() throws IOException {
		if (this.closed) {
			throw new IOException("The database has been already closed!");
		}
		this.referencesMetadata.close();
		this.fileIO.close();
		this.closed = true;
	}

	@Override
	public boolean isClosed() {
		return closed;
	}

	@Override
	public void closeAndClean() throws IOException {
		if (!this.closed) {
			this.close();
		}
		Path newDataFile = dataFile.resolveSibling("compressed-data-file.tmp");
		Path newReferencesFile = referencesMetaFile.resolveSibling("compressed-references-file.tmp");
		Path backupDataFile = dataFile.resolveSibling("backup-data.db.bak");
		Path backupReferencesFile = referencesMetaFile.resolveSibling("backup-references.dat.bak");
		Files.copy(dataFile, backupDataFile, StandardCopyOption.REPLACE_EXISTING, StandardCopyOption.COPY_ATTRIBUTES);
		Files.copy(referencesMetaFile, backupReferencesFile, StandardCopyOption.REPLACE_EXISTING, StandardCopyOption.COPY_ATTRIBUTES);
		Files.move(dataFile, newDataFile, StandardCopyOption.REPLACE_EXISTING);
		Files.move(referencesMetaFile, newReferencesFile, StandardCopyOption.REPLACE_EXISTING);
		DatabaseCore databaseToClean = new DatabaseCore(newDataFile, newReferencesFile);
		DatabaseCore newDatabase = new DatabaseCore(dataFile, referencesMetaFile);

		long referencesCount = databaseToClean.referencesMetadata.getFirstFreeReference();
		long writtenReferences = 0;

		for (int referenceID = 0; referenceID < referencesCount; referenceID++) {
			try {
				ReferenceInfo ref = databaseToClean.referencesMetadata.getReferenceInfo(referenceID);
				if (!NONEXISTENT_REFERENCE_INFO.equals(ref)) {
					ByteBuffer buffer = databaseToClean.referencesIO.readFromReference(referenceID);
					newDatabase.referencesIO.writeToReference(referenceID, ref.getCleanerId(), buffer.limit(), buffer);
					writtenReferences++;
				}
			} catch (IOException ex) {
				System.out.println("Error while reading reference " + referenceID + ". References written: " + writtenReferences);
			}
		}
		System.out.println("[Core Cleaner] References written: " + writtenReferences + ". Removed " + (referencesCount - writtenReferences) + " references.");
		databaseToClean.close();
		newDatabase.close();
		Files.deleteIfExists(newDataFile);
		Files.deleteIfExists(newReferencesFile);
	}
}
